/*
 * Copyright 2010 Inalogic Inc.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3, as
 * published by the  Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the applicable version of the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of both the GNU Lesser General Public
 * License version 3 along with this program.  If not, see
 * <http://www.gnu.org/licenses/>
 *
 * Authored by: Jay Taoko <jaytaoko@inalogic.com>
 *
 */


#ifndef STACKMANAGER_H
#define STACKMANAGER_H

#include "BaseWindow.h"

namespace nux
{

  class MenuPage;
  class PBuffer;
  class WindowThread;
  class View;
  class InputArea;
  class Area;
  class BaseWindow;
  class PaintLayer;

  class WindowCompositor
  {
  public:
    WindowCompositor();
    ~WindowCompositor();


    //! Get the Geometry of the tooltip based on the BaseWindow that initiated it.
    Geometry GetTooltipGeometry() const;

    //! Get the Geometry of the tooltip based on the MainWindow.
    Geometry GetTooltipMainWindowGeometry() const;

    bool MouseDown (Point pt);

    bool MouseMove (Point pt);
    bool MouseUp (Point pt);

    void ProcessEvent (IEvent &ievent);


    ObjectPtr<IOpenGLFrameBufferObject>& GetWindowFrameBufferObject()
    {
      return m_FrameBufferObject;
    }
    ObjectPtr<IOpenGLFrameBufferObject> m_FrameBufferObject;

    ObjectPtr< IOpenGLBaseTexture > GetScreenBlurTexture();

    void StartModalWindow (ObjectWeakPtr<BaseWindow>);
    void StopModalWindow (ObjectWeakPtr<BaseWindow>);

    void AddMenu(MenuPage* menu, BaseWindow *window, bool OverrideCurrentMenuChain = true);
    void RemoveMenu(MenuPage* menu);
    void CleanMenu();

    void PushModalWindow (ObjectWeakPtr<BaseWindow> window);

    void SetWidgetDrawingOverlay (InputArea *ic, BaseWindow *OverlayWindow);
    InputArea *GetWidgetDrawingOverlay();

    void SetTooltip (InputArea *TooltipArea, const TCHAR *TooltipText, int x, int y);
    /*!
        Return true if the mouse is still inside the area that initiated the tooltip;

        @param x    The mouse x coordinate on screen.
        @param y    The mouse y coordinate on screen.
        @return     Return true is the mouse is still inside the area.
    */
    bool ValidateMouseInsideTooltipArea (int x, int y);
    /*!
        Return true if there is a valid tooltip active.

        @return     Return true if there is a valid tooltip active.
    */
    bool IsTooltipActive();
    void CancelTooltip();

    void SetAreaEventRoot (int x, int y)
    {
      m_EventRoot.Set (x, y);
    }

    const IEvent *GetCurrentEvent() const
    {
      return m_CurrentEvent;
    }
    long ProcessEventOnObject (IEvent &ievent, Area *object, long TraverseInfo, long ProcessEventInfo);

    void SetBackgroundPaintLayer (AbstractPaintLayer *bkg);

    /*!
        A special BaseWindow that is always on top of all other BaseWindow. It is even above the BaseWindow that is selected.
        \sa m_SelectedWindow, \sa GetSelectedWindow.
    */
    void SetAlwaysOnFrontWindow (BaseWindow *window);


    //! Enable the exclusive event input mode.
    /*!
        Set the exclusive event input area (\sa _exclusive_input_area). The greedy input area gets all input events (mouse and keyboard).
        The exclusive input mode can only be set if there is no exclusive input area already set.
        To disable the exclusive input move, call DisableExclusiveInputArea with the current exclusive input area as parameter.
        The exclusive event input mode can only change once during the processing of one event. The change it again, 
        you have to wait for the next event cycle.
        \sa DisableExclusiveInputArea.
        @return True, if the exclusive input mode was enabled.
    */
    bool EnableExclusiveInputArea (InputArea *input_area);
    
    //! Disable the exclusive event input mode.
    /*!
        Disable the exclusive event input mode. It can only be disable if the current exclusive input area is passed as parameter.
        \sa EnableExclusiveInputArea.
        @return True, if the exclusive input mode was disabled.
    */
    bool DisableExclusiveInputArea (InputArea *input_area);

    //! Return true if the system is in exclusive input event mode.
    /*!
        @return True if the system is in exclusive input mode.
    */
    bool InExclusiveInputMode ();

    //! Set the rendering surface for the current rendering.
    /*!
        This function is used to restore the rendering surface according to the system state. This is necessary after using a custom frame buffer object.
    */
    void RestoreRenderingSurface ();

  private:
    //! Render the interface.
    void Draw (bool SizeConfigurationEvent, bool force_draw);

    void DrawPopup (bool force_draw);
    void DrawMenu (bool force_draw);
    void DrawOverlay (bool force_draw);
    void DrawTooltip (bool force_draw);

    //! Render all top views.
    /*!
        @force_draw True if the entire surface of the backup rendering mush flushed.
        @WindowList The list of top views.
        @draw_modal True if the top view that is modal is to be rendered.
        @use_fbo True if TopViews should be backed by an fbo.
    */
    void RenderTopViews (bool force_draw, std::list< ObjectWeakPtr<BaseWindow> >& WindowList, bool draw_modal, bool use_fbo);

    //! Render the content of a top view.
    void RenderTopViewContent (BaseWindow *window, bool force_draw);

    void RenderMainWindowComposition (bool force_draw, bool UseFBO);



    /*!
        Render a textured quad the quad has the size of the texture. The texture maybe the main window texture or a BaseWindow texture.

        @param HWTexture            Texture to render.
        @param x                    Coordinates of the top left corner of the quad.
        @param y                    Coordinates of the top left corner of the quad.
        @param RenderToMainTexture  If true, render to the main window texture. If false, render to the default back buffer.
        @param BluredBackground     If true, the texture is blended with the blurred version of the main window texture.
    */
    void PresentBufferToScreen (ObjectPtr<IOpenGLBaseTexture> HWTexture, int x, int y, bool RenderToMainTexture, bool BluredBackground = false);
    void PresentRendering();

    /*!
        Set the main color render target as the texture to draw into.
    */
    void SetMainColorRT();

    /*!
        Draw a Texture into the main color render target.
        @param x    Destination coordinates.
        @param y    Destination coordinates.
    */
    void CopyTextureToMainColorRT (ObjectPtr<IOpenGLBaseTexture> HWTexture, int x, int y);

    /*!
        Set the composition render target as the texture to draw into.
    */
    void SetCompositionRT();

    /*!
        Draw a Texture into the composition render target.
        @param x    Destination coordinates.
        @param y    Destination coordinates.
    */
    void CopyTextureToCompositionRT (ObjectPtr<IOpenGLBaseTexture> HWTexture, int x, int y);

    //! Push a floating view just above another floating view.
    /*!
        Note that only the top_floating_view is moving. The overall position of the reference view is not changing. 
        @param top_floating_view The view to place above bottom_floating_view.
        @param top_floating_view The reference view that will be below top_floating_view.
        @param strict If true and top_floating_view is already above bottom_floating_view, then bring top_floating_view lower
        so that it is strictly above bottom_floating_view.
    */
    void PushHigher (BaseWindow *top_floating_view, BaseWindow *bottom_floating_view, bool strict = false);
    //! Push a floating view at the top of the stack.
    void PushToFront (BaseWindow *bottom_floating_view);
    //! Push a floating view at the bottom of the stack.
    void PushToBack (BaseWindow *bottom_floating_view);

    /*!
        Returns the BaseWindow that is at the top of the BaseWindow stack, excluding the BaseWindow that is
        chosen to be always on to.
        \sa m_SelectedWindow. \sa SetAlwaysOnFrontWindow
    */
    BaseWindow *GetSelectedWindow();

    
    BaseWindow *GetFocusAreaWindow()
    {
      return m_FocusAreaWindow.GetPointer ();
    }

    //! Set the top view that is about to be processed (event or rendering).
    /*!
        Before event processing or rendering, this should be called to set the TopView that is about 
        to be processed. This function is used internally by the system.
    */
    void SetProcessingTopView (BaseWindow* window)
    {
      m_CurrentWindow = window;
    }

    //! Get the top view that is being processed (event or rendering).
    /*!
        Get the active TopView during and event processing or rendering.
    */
    BaseWindow *GetProcessingTopView()
    {
      return m_CurrentWindow.GetPointer ();
    }

    void SetFocusAreaWindow (BaseWindow *window)
    {
      m_FocusAreaWindow = window;
    }

    void SetCurrentEvent (IEvent *event)
    {
      m_CurrentEvent = event;
    }

    void EnsureAlwaysOnFrontWindow ();

    void FormatRenderTargets (int width, int height);

    //void UpdatePostProcessRT();

    /*!
        Floating Area need to be informed when the main window has been resized.

        @param Width    New width of the window.
        @param Height   New height of the window.
    */
    void FloatingAreaConfigureNotify(int Width, int Height);

    void SetMouseFocusArea (InputArea *area);
    InputArea *GetMouseFocusArea();
    void SetMouseOverArea (InputArea *area);
    InputArea *GetMouseOverArea();
    void SetPreviousMouseOverArea (InputArea *area);
    InputArea *GetPreviousMouseOverArea();

    void RegisterWindow (BaseWindow*);
    void UnRegisterWindow (BaseWindow*);

    // We use Rectangle texture to attach to the framebuffer because some GPU like the Geforce FX 5600 do not
    // have support for ARB_texture_non_power_of_two. However it does support ARB_texture_recatangle.
    struct RenderTargetTextures
    {
      ObjectPtr<IOpenGLBaseTexture> color_rt;
      ObjectPtr<IOpenGLBaseTexture> depth_rt;
    };
    ObjectPtr<IOpenGLBaseTexture> m_MainColorRT;
    ObjectPtr<IOpenGLBaseTexture> m_MainDepthRT;
    ObjectPtr<IOpenGLBaseTexture> m_CompositionRT;

    //! Return the RenderTargetTextures structure of a TopView.
    RenderTargetTextures &GetWindowBuffer (BaseWindow* window);

    ObjectWeakPtr<BaseWindow> m_CurrentWindow;    //!< BaseWindow where event processing or rendering is happening.
    ObjectWeakPtr<BaseWindow> m_FocusAreaWindow;  //!< The BaseWindow that contains the m_MouseFocusArea.
    ObjectWeakPtr<BaseWindow> m_MenuWindow;       //!< The BaseWindow that owns the menu being displayed;
    IEvent* m_CurrentEvent; 

    /*!
        The area that has the mouse focus. Mouse focus happens when the mouse down is received on and area.
        The mouse focus stops when the last mouse button is released.
        Incidentally, when the mouse focus first occurs on an area, we set the keyboard focus over that area.
        The keyboard focus of an area stops when the mouse down happens over a different area.
    */
    InputArea* m_MouseFocusArea;     
    InputArea* m_MouseOverArea;      //!< The base area that has the mouse directly over itself.
    InputArea* m_PreviousMouseOverArea;

    //! Set the exclusive input area according to _pending_exclusive_input_mode_action.
    /*!
        Following the event processing cycle, it is necessary to setup the exclusive input area is _pending_exclusive_input_mode_action is true.
        The exclusive input area status always takes effect after the event processing cycle.
    */
    void ExecPendingExclusiveInputAreaAction ();

    //! Get the input area that has the exclusivity on events.
    /*!
        @return The input area that has the exclusivity on all events.
    */
    InputArea *GetExclusiveInputArea ();

    /*!
        The exclusive input area gets all events without exception (greedy). The exclusive input area may decide to pass events 
        down to other areas. If it does, the following restrictions apply:
          - The other input area cannot have the mouse focus.
          - They cannot have the keyboard focus.
          - they cannot call ForceStartFocus or ForceStopFocus.
          - No synthetic events: 
            * mouse click
            * mouse drag
          - Only atomic events:
            * mouse down
            * mouse up
            * mouse move
            * mouse enter/leave
    */
    InputArea* _exclusive_input_area;
    
    /*!
        \a _exclusive_input_area is true when there is an active greedy input area.
    */
    bool _in_exclusive_input_mode;

    /*!
        The exclusive input mode starts after after events have been processed inside ProcessEvent ().
        This flags signals that some action are required to enable/disable the exclusive input mode.
    */
    bool _pending_exclusive_input_mode_action;

    //! True while events are being processed inside ProcessEvent ().
    bool _inside_event_processing;

    //! True while rendering is being done.
    bool _inside_rendering_cycle;

    InputArea* OverlayDrawingCommand;
    ObjectWeakPtr<BaseWindow> m_OverlayWindow;            //!< The window that owns the overlay;
    ObjectWeakPtr<BaseWindow> _tooltip_window;            //!< The window that owns the tooltip;
    Geometry    _tooltip_geometry;              //!< The geometry of the entire tooltip It includes the decoration surrounding the text such as round corners.
    Geometry    _tooltip_mainwindow_geometry;   //!< Same as _tooltip_geometry but based on the entire physical window of the application.
    Geometry    _tooltip_text_geometry;         //!< The geometry of the text area of the tooltip.
    Point m_EventRoot;


    AbstractPaintLayer *m_Background;

    std::list< ObjectWeakPtr<BaseWindow> > m_WindowList;
    std::list< ObjectWeakPtr<BaseWindow> > m_ModalWindowList;
    ObjectWeakPtr<BaseWindow>            _always_on_front_window;  //!< Floating view that always remains on top.

    std::list<MenuPage *> *m_MenuList;

    /*!
        The BaseWindow where the last mouse down event happened. 
        This BaseWindow will be raised to the top of the stack. 
        \sa GetSelectedWindow.
    */
    ObjectWeakPtr<BaseWindow> m_SelectedWindow;

    std::map< BaseWindow*, struct RenderTargetTextures > m_WindowToTextureMap;

    ObjectWeakPtr<BaseWindow> m_ModalWindow;
    Point m_MouseLastPos;
    Point m_MouseCurrentPos;

    bool m_PopupRemoved;
    bool m_MenuRemoved;
    bool m_MouseUp;
    bool m_MouseDown;

    // Window Geometry
    int m_Width;
    int m_Height;

    NString m_TooltipText;
    InputArea *m_TooltipArea;
    int m_TooltipX;
    int m_TooltipY;

//     bool m_FullSceneBlurUpdated;
//     ObjectPtr<IOpenGLBaseTexture> m_BlurTexture;
//     ObjectPtr<IOpenGLBaseTexture> m_FullSceneMip0;
//     ObjectPtr<IOpenGLBaseTexture> m_FullSceneMip1;
//     ObjectPtr<IOpenGLBaseTexture> m_FullSceneMip2;

  private:
    WindowCompositor (const WindowCompositor &);
    // Does not make sense for a singleton. This is a self assignment.
    WindowCompositor &operator= (const WindowCompositor &);
    // Declare operator address-of as private
    WindowCompositor *operator &();

    friend class InputArea;
    friend class WindowThread;
    friend class TimerHandler;
    friend class MenuBar;
    friend class MenuPage;
    friend class BaseWindow;
    friend class HSplitter;
    friend class VSplitter;
    friend class TableCtrl;
    friend class View;
  };

}
#endif // STACKMANAGER_H

